/*
 * Title:        GridSim Toolkit
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation
 *               of Parallel and Distributed Systems such as Clusters and Grids
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 *
 * $Id: GridResource.java,v 1.14 2003/06/30 05:21:01 anthony Exp $
 */

package gridsim;

import java.util.*;
import java.io.*;
import eduni.simjava.*;


/**
 * GridSim GridResource extends the <b>gridsim.GridSim</b> class and
 * gains communication and
 * concurrent entity capability. An instance of this class stimulates a resource
 * with properties defined in an object of
 * <b>gridsim.ResourceCharacteristics</b> class.
 * <p>
 * The process of creating a Grid resource is as follows:
 * <ol>
 * <li> create PE (Processing Element) objects with a suitable MIPS (Million
 *      Instructions Per Second) or SPEC (Standard Performance Evaluation
 *      Corporation) rating;
 * <li> assemble them together to create a machine;
 * <li> group one or more objects of the machine to form a resource
 * </ol>
 * <p>
 * A resource having a single machine with one or more PEs (Processing Elements)
 * is managed as a time-shared system using a round-robin scheduling algorithm.
 * A resource with multiple machines is treated as a distributed memory cluster
 * and is managed as a space-shared system using FCFS (First Come Firt Serve)
 * scheduling policy or its variants.
 *
 * @author       Manzur Murshed and Rajkumar Buyya
 * @version      2.1, June 2003
 * @see gridsim.GridSim
 * @see gridsim.ResourceCharacteristics
 * @invariant $none
 */
public class GridResource extends GridSim
{
    // |Machines| > 1 is cluster managed by queueing system
    private ResourceCharacteristics resource_;
    private ResourceCalendar resCalendar_;

    // Gridlets in Queued for execution: used in shace-shared resources.
    private ResGridletList gridletQueueList_;
    private ResGridletList gridletInExecList_;
    private ResGridletList gridletFinishList_;  // for accounting

    private double lastUpdateTime_;
    private double localLoad_;
    private int scheduleId_;
    private Accumulator accTotalLoad_;


    /**
     * Allocates a new GridResource object
     * @param name       the name to be associated with this entity (as
     *                   required by Sim_entity class from simjava package)
     * @param baud_rate  communication speed
     * @param seed       the initial seed
     * @param resource   an object of ResourceCharacteristics
     * @param peakLoad   the load during peak times
     * @param offpeakLoad   the load during off peak times
     * @param relativeHolidayLoad   the load during holiday times
     * @param weekends   a linked-list contains the weekend days
     * @param holidays   a linked-list contains the public holidays
     * @throws Exception This happens when creating this entity before
     *                   initializing GridSim package or this entity name is
     *                   <tt>null</tt> or empty
     * @see gridsim.GridSim#Init(int, Calendar, boolean, String[], String[],
     *          String)
     * @pre $none
     * @post $none
     */
    public GridResource(String name, double baud_rate, long seed,
            ResourceCharacteristics resource, double peakLoad,
            double offpeakLoad, double relativeHolidayLoad,
            LinkedList weekends, LinkedList holidays) throws Exception
    {
        super(name, baud_rate);
        this.resource_ = resource;

        this.resource_.setResourceID( super.get_id() );
        this.resCalendar_ = new ResourceCalendar(
                resource_.getResourceTimeZone(), peakLoad, offpeakLoad,
                relativeHolidayLoad, weekends, holidays, seed);

        this.gridletQueueList_ = new ResGridletList();
        this.gridletInExecList_ = new ResGridletList();
        this.gridletFinishList_ = new ResGridletList();
        this.lastUpdateTime_ = GridSim.clock();
        this.localLoad_ = resCalendar_.getCurrentLoad();
        this.scheduleId_ = 0;
        this.accTotalLoad_ = new Accumulator();
        this.accTotalLoad_.add( totalLoad() );
    }

    /**
     * A method that registers a resource to be scheduled, then processes
     * internal and external events.
     * <p>
     * The services available to other GridSim entities are:
     * <ul>
     *      <li> GridSimTags.RESOURCE_CHARACTERISTICS </li>
     *      <li> GridSimTags.RESOURCE_DYNAMICS </li>
     *           supporting for the following allocation policy:
     *           <ol>
     *              <li> ResourceCharacteristics.TIME_SHARED </li>
     *              <li> ResourceCharacteristics.SPACE_SHARED </li>
     *           </ol>
     *      <br>
     *      <li> GridSimTags.GRIDLET_SUBMIT </li>
     *           supporting for the following allocation policy:
     *           <ol>
     *              <li> ResourceCharacteristics.TIME_SHARED </li>
     *              <li> ResourceCharacteristics.SPACE_SHARED </li>
     *           </ol>
     * </ul>
     * @pre $none
     * @post $none
     */
    public void body()
    {
        // RESOURCE REGISTRATION:
        super.send(output, GridSimTags.SCHEDULE_NOW,
                GridSimTags.REGISTER_RESOURCE,
                new IO_data( new Integer(super.get_id()), 4,
                    GridSim.getGridInfoServiceEntityId() ));

        // Place Hourly Interrupt Request to itself: This is handle change in
        // load/access cost every hour (assumption)
        Calendar NextHour = resCalendar_.getCurrentCalendar();
        NextHour.add(Calendar.HOUR, 1);
        NextHour.set(Calendar.MINUTE, 0);
        NextHour.set(Calendar.SECOND, 0);

        // if delay is negative, then it is unreasonable. So resets to 0.0
        double delay = resCalendar_.getSimulationTime(NextHour)-GridSim.clock();
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.send(super.get_id(), delay, GridSimTags.INSIGNIFICANT);

        // Process Events until END_OF_SIMULATION is received from the
        // GridSimShutdown Entity
        Sim_event ev = new Sim_event();
        for ( sim_get_next(ev); ev.get_tag() != GridSimTags.END_OF_SIMULATION;
                sim_get_next(ev) )
        {
            // Internal Event if the event source is this entity
            if ( ev.get_src() == super.get_id() ) {
                processInternalEvents(ev);
            }

            // External Events
            else {
                processExternalEvents(ev);
            }
        }

        // CHECK for ANY INTERNAL EVENTS WAITING TO BE PROCESSED
        while (super.sim_waiting() > 0)
        {
            // wait for event and ignore since it is likely to be related to
            // internal event scheduled to simulate PE allocation to Gridlets
            System.out.println("-->GridSim detected Internal Events in " +
                    "GridResource entity and they are ignored!");

            sim_get_next(ev);
        }

        // remove I/O entities created during construction of this entity
        super.terminateIOEntities();
    }

    /**
     * Allocates one of the PEs to Gridlet for execution and schedules
     * an internal event to be delivered at completion time. This is done by:
     * <ol>
     * <li> Identifing Unallocated PEs;
     * <li> If some free PEs are available, map gridlet to available PE by:
     *    <ul>
     *    <li> Total Execution Time = [GridletMI / (PE's MPIS Rating)] + IO Time
     *    <li> Set an Event to be se sent after Total_Execution_Time
     *         indicating completion of Gridlet.
     *         <br>
     *         <b>Else</b>, place Gridlet into "Job Queue".
     *    </ul>
     * </ol>
     *
     * @param gl a Gridlet to be processed
     * @deprecated As of GridSim 2.1, replaced by
     *             {@link #spaceShared_AllocatePEtoGridlet(Gridlet)}
     * @pre gl != null
     * @post $none
     */
    public void SpaceShare_AllocatePEtoGridlet(Gridlet gl) {
        this.spaceShared_AllocatePEtoGridlet(gl);
    }

    /**
     * Allocates one of the PEs to Gridlet for execution and schedules
     * an internal event to be delivered at completion time. This is done by:
     * <ol>
     * <li> Identifing Unallocated PEs;
     * <li> If some free PEs are available, map gridlet to available PE by:
     *    <ul>
     *    <li> Total Execution Time = [GridletMI / (PE's MPIS Rating)] + IO Time
     *    <li> Set an Event to be se sent after Total_Execution_Time
     *         indicating completion of Gridlet.
     *         <br>
     *         <b>Else</b>, place Gridlet into "Job Queue".
     *    </ul>
     * </ol>
     *
     * @param gl a Gridlet to be processed
     * @pre gl != null
     * @post $none
     */
    public void spaceShared_AllocatePEtoGridlet(Gridlet gl)
    {
        // IDENTIFY MACHINE which has FREEE PE and add Gridlet to it.
        MachineList MyMachineList = resource_.getMachineList();
        Machine myMachine = MyMachineList.getMachineWithFreePE();

        if (myMachine == null)
        {
            System.out.println("GridResource Error: There is no machine " +
                    "with free PE for space-shared scheduling.");

            System.out.println("The caller of this rountine is assumed " +
                    "to ensure that there is atleast one PE free.");
            return;
        }

        PEList MyPEList = myMachine.getPEList();
        int FreePE_ID = MyPEList.getFreePEID();

        // ALLOCATE IMMEDIATELY
        try {
            gl.setGridletStatus(Gridlet.INEXEC);
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        gl.setExecStartTime( GridSim.clock() );

        ResGridlet rgl = new ResGridlet(gl);
        rgl.setMachineAndPEID(myMachine.getMachineID(), FreePE_ID);
        gridletInExecList_.add(rgl);

        // Set allocated PE to BUSY status
        PE myPE = (PE) MyPEList.get(FreePE_ID);
        myPE.setStatusBusy();

        // Identify Completion Time and Set Interrupt
        rgl.setFinishTimeForSpaceSharedResource(
                MyPEList.getMIPSRating(FreePE_ID) );

        scheduleId_++;
        
        // if delay is -ve, then it is unreasonable. So resets to 0.0
        double delay = rgl.getGridletFinishTime() - GridSim.clock();
        if (delay < 0.0) {
            delay = 0.0;
        }
        
        super.send(super.get_id(), delay, scheduleId_, rgl);
    }


    ///////////////////// INTERNAL CLASS //////////////////////

    /**
     * Gridlets MI share in Time Shared Mode
     */
    private class MIShares
    {
        public double max; // max amount of MI share Gridlets can get
        public double min; // min amount of MI share Gridlets can get when

        // it is executed on a PE that runs one extra Gridlet
        public int maxCount; // Total number of Gridlets that get Max share.
    }


    ////////////////////// PRIVATE METHODS //////////////////////////////

    /**
     * Processes external events which signify the completion of Gridlet
     * @param ev    a Sim_event object
     * @pre ev != null
     * @post $none
     */
    private void processExternalEvents(Sim_event ev)
    {
        int src_id = -1;
        switch ( ev.get_tag() )
        {
            // Resource characteristics inquiry
            case GridSimTags.RESOURCE_CHARACTERISTICS:
                src_id = ( (Integer) ev.get_data() ).intValue();
                super.send(output, GridSimTags.SCHEDULE_NOW, ev.get_tag(),
                        new IO_data(resource_,resource_.getByteSize(),src_id));
                break;

            // Resource dynamic info inquiry:
            case GridSimTags.RESOURCE_DYNAMICS:
                src_id = ( (Integer) ev.get_data() ).intValue();
                switch ( resource_.getResourceAllocationPolicy() )
                {
                    case ResourceCharacteristics.TIME_SHARED:
                        super.send( output, GridSimTags.SCHEDULE_NOW,
                                ev.get_tag(), new IO_data(accTotalLoad_,
                                    Accumulator.getByteSize(), src_id) );
                        break;

                    // Inform number of PE available/free.:
                    case ResourceCharacteristics.SPACE_SHARED:
                        super.send( output, GridSimTags.SCHEDULE_NOW,
                                ev.get_tag(), new IO_data(accTotalLoad_,
                                    Accumulator.getByteSize(), src_id) );
                        break;

                    default:
                        System.out.println("GridResource.body() : Unable to " +
                                "handle request from ResourceCharacteristics " +
                                "with constant number " +
                                resource_.getResourceAllocationPolicy() );
                        break;
                }
                break;

            // New gridlet arrives
            case GridSimTags.GRIDLET_SUBMIT:
                switch (resource_.getResourceAllocationPolicy())
                {
                    case ResourceCharacteristics.TIME_SHARED:
                    case ResourceCharacteristics.SPACE_SHARED:
                        gridletArrives( (Gridlet) ev.get_data() );
                        break;

                    default:
                        System.out.println("GridResource.body() : Unable to " +
                                "handle request from ResourceCharacteristics " +
                                "with constant number " +
                                resource_.getResourceAllocationPolicy() );
                        break;
                }
                break;

            default:
                System.out.println("GridResource.body() : Unable to " +
                        "handle request from GridSimTags " +
                        "with constant number " +
                        ev.get_tag() );
                break;
        }
    }

    /**
     * Processes internal events which signify the completion of Gridlet
     * @param ev    a Sim_event object
     * @pre ev != null
     * @post $none
     */
    private void processInternalEvents(Sim_event ev)
    {
        if (ev.get_data() != null) // means, the ev.get_data() returns Gridlet
        {
            switch ( resource_.getResourceAllocationPolicy() )
            {
                // A Gridlet is finished if event received is the most
                // recently scheduled one
                // Otherwise, Ignore the Event since situation changed
                // because of an arrival new Gridlet
                // which might have changed the completion time.
                case ResourceCharacteristics.TIME_SHARED:
                    // event is the most recently scheduled one ?
                    if (ev.get_tag() == scheduleId_) {
                        gridletFinishes( (ResGridlet) ev.get_data() );
                    }
                    break;

                // Unlike Space shared resource, once scheduled
                // Gridlet to a PE, it is dedicated to
                // it until it finishes.
                case ResourceCharacteristics.SPACE_SHARED:
                    gridletFinishes( (ResGridlet) ev.get_data() );
                    break;

                default:
                    System.out.println("GridResource.body() : Unable to " +
                        "handle request from ResourceCharacteristics " +
                        "with constant number " +
                        resource_.getResourceAllocationPolicy() );
                    break;
            }
        }

        else {
            hourlyInterrupt();
        }
    }

    /**
     * Gets the total load of a resource
     * @return total load
     * @pre $none
     * @post $result >= 0.0
     */
    private double totalLoad()
    {
        int NoOfGridletsPerPE = (int) Math.ceil( (gridletInExecList_.size()
                    + 1.0) / resource_.getNumPE() );

        return 1.0 - (1 - localLoad_) / NoOfGridletsPerPE;
    }

    /**
     * Identifies MI share (max and min) each Gridlet gets for
     * a given TimeSpan
     * @param TimeSpan duration
     * @return  the total MI share that a Gridlet gets for a given
     *          <tt>TimeSpan</tt>
     */
    private MIShares getMIShare(double TimeSpan)
    {
        MIShares obj = new MIShares();

        // 1 - localLoad_ = available MI share percentage
        double TotalMIperPE = resource_.getMIPSRatingOfOnePE() * TimeSpan
                                    * (1 - localLoad_);

        int NoOfGridletsInExec = gridletInExecList_.size();
        int NoOfPEs = resource_.getNumPE();

        // minimum number of Gridlets that each PE runs.
        int glDIVpe = NoOfGridletsInExec / NoOfPEs;

        // number of PEs that run one extra Gridlet
        int glMODpe = NoOfGridletsInExec % NoOfPEs;

        // If NoOfGridletsInExec > NoOfPEs then, divide MIShare among
        // all Gridlets
        if (glDIVpe>0)
        {
            // this is for PEs that run one extra Gridlet
            obj.min = TotalMIperPE / (glDIVpe + 1);
            obj.max = TotalMIperPE / (glDIVpe);
            obj.maxCount = (NoOfPEs - glMODpe) * glDIVpe;
        }
        else
        {
            // All Gridlets are executed on dedicated PE since
            // NoOfGridletsInExec < NoOfPEs
            // full PE share: i.e PE is dedicated for executing a single Gridlet
            obj.max = TotalMIperPE;
            obj.min = obj.max;  // this is for a PE that runs one extra Gridlet
            obj.maxCount = NoOfGridletsInExec;
        }

        return obj;
    }

    /**
     * Updates amount of MIs consumed by every Gridlet from the last update.
     * This method need to be invoked for events
     * (gridlet arrival, finish, hourly interrupt)
     * @pre $none
     * @post $none
     */
    private void timeShared_UpdateGridletProcessingSoFar()
    {
        // Identify MI share for the duration (from last event time)
        double timeSpan = GridSim.clock() - lastUpdateTime_;
        MIShares shares = getMIShare(timeSpan);

        // allocate MI share accordingly
        ResGridlet obj;
        for (int i = 0; i < gridletInExecList_.size(); i++)
        {
            obj = (ResGridlet) gridletInExecList_.get(i);
            if (i < shares.maxCount) {
                obj.updateGridletFinishedSoFar(shares.max);
            }
            else {
                obj.updateGridletFinishedSoFar(shares.min);
            }
        }

        // Update Current Time as Last Update
        lastUpdateTime_ = GridSim.clock();
        localLoad_ = resCalendar_.getCurrentLoad();
        accTotalLoad_.add( totalLoad() );
    }

    /**
     * Forecasts Finish Time of all Gridlets
     * @pre $none
     * @post $none
     */
    private void timeShared_ForecastFinishTimeOfGridlet()
    {
        // Identify MIPS share for each Gridlet considering
        // current Gridlets + No of PEs.
        double timeSpan = 1; // One second
        MIShares MIPSshare_per_gridlet = getMIShare(timeSpan);

        ResGridlet obj;
        for (int i = 0; i < gridletInExecList_.size(); i++)
        {
            obj = (ResGridlet) gridletInExecList_.get(i);
            if (i < MIPSshare_per_gridlet.maxCount)
            {
                obj.setFinishTimeForTimeSharedResource(
                        MIPSshare_per_gridlet.max, resCalendar_ );
            }
            else
            {
                obj.setFinishTimeForTimeSharedResource(
                        MIPSshare_per_gridlet.min, resCalendar_ );
            }
        }
    }

    /**
     * Identifies time by which one of the Gridlet finishes and sets that time
     * as a time for raising the next event
     * @pre $none
     * @post $none
     */
    private void timeShared_ForecastAndScheduleNextCompletionEvent()
    {
        timeShared_ForecastFinishTimeOfGridlet();

        // FOR TIME SHARED: Identify a Gridlet which has the smallest
        // remaining completion time
        ResGridlet MinGl = null;
        ResGridlet gl = null;

        double finishTime = 0.0;
        for (int i = 0; i < gridletInExecList_.size(); i++)
        {
            gl = (ResGridlet) gridletInExecList_.get(i);
            finishTime = gl.getGridletFinishTime();

            // Cannot finish in this hour slot
            if (finishTime < 0.0)  {
                continue;
            }

            if (MinGl == null) {
                MinGl = gl;   //assume first small Gridlet
            }
            else if ( finishTime < MinGl.getGridletFinishTime() ) {
                MinGl = gl;   // smallest gridlet identified so far
            }
        }

        // Schedule Gridlet completion event
        if (MinGl != null)
        {
            scheduleId_++;

            // if delay is -ve, then it is unreasonable. So resets to 0.0
            double delay = MinGl.getGridletFinishTime() - GridSim.clock();
            if (delay < 0.0) {
                delay = 0.0;
            }

            super.send(super.get_id(), delay, scheduleId_, MinGl);
                    
        }
    }

    /**
     * Processes an event related to Gridlet arrival
     * for Time-Shared Resources. This is done by:
     * <ol>
     *      <li> updating processing of Gridlets so far;
     *      <li> adding arrived jobs to gridletInExecList_ queue; and
     *      <li> updating next event related to the gridlet processing
     * </ol>
     *
     * @param gl a Gridlet object to be processed
     * @pre gl != null
     * @post $none
     */
    private void gridletArrives(Gridlet gl)
    {
        ResGridlet rgl;

        try
        {
            switch ( resource_.getResourceAllocationPolicy() )
            {
                case ResourceCharacteristics.TIME_SHARED:
                    timeShared_UpdateGridletProcessingSoFar();

                    // add Gridlets to Execution Queue
                    gl.setGridletStatus(Gridlet.INEXEC);

                    // set Gridlet Submission and StartTime to the same time.
                    gl.setSubmissionTime( GridSim.clock() );
                    gl.setExecStartTime( GridSim.clock() );
                    rgl = new ResGridlet(gl);
                    gridletInExecList_.add(rgl);

                    // schedule next event
                    timeShared_ForecastAndScheduleNextCompletionEvent();
                    break;

                case ResourceCharacteristics.SPACE_SHARED:
                    gl.setGridletStatus(Gridlet.QUEUED);

                    // Set Submission Time only.
                    gl.setSubmissionTime(GridSim.clock());
                    if ( gridletInExecList_.size() < resource_.getNumPE() ) {
                        spaceShared_AllocatePEtoGridlet(gl);
                    }
                    else
                    {
                        // ADD GRIDLET To QUEUE.
                        gl.setGridletStatus(Gridlet.QUEUED);
                        gl.setExecStartTime( GridSim.clock() );
                        rgl = new ResGridlet(gl);
                        gridletQueueList_.add(rgl);
                    }
                    break;

                default:
                    System.out.println("GridResource.body() : Unable to " +
                            "handle request from ResourceCharacteristics " +
                            "with constant number " +
                            resource_.getResourceAllocationPolicy() );
                    break;
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Processes an event related to the Gridlet that just finished
     * processing, by:
     * <ol>
     *   <li> sending back the finished Gridlet back to requester (user/broker)
     *   <li> updating gridletInExecList_ and gridletFinishList_ record
     *   <li> updating events related to PE sharing
     * </ol>
     *
     * @param rgl a ResGridlet object
     * @pre rgl != null
     * @post $none
     */
    private void gridletFinishes(ResGridlet rgl)
    {
        timeShared_UpdateGridletProcessingSoFar();
        try {
            rgl.getGridlet().setGridletStatus(Gridlet.SUCCESS);
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        double WallClockTime = GridSim.clock() - rgl.getGridletArrivalTime();
        double CPUTime = 0;
        switch ( resource_.getResourceAllocationPolicy() )
        {
            // Only CPU time consumed is measured since during I/O time,
            // CPU might have been used by other Gridlets
            case ResourceCharacteristics.TIME_SHARED:
                CPUTime = rgl.getGridlet().getGridletLength() /
                                resource_.getMIPSRatingOfOnePE();
                break;

            case ResourceCharacteristics.SPACE_SHARED:
                CPUTime = GridSim.clock() - rgl.getGridlet().getExecStartTime();
                break;

            default:
                break;
        }

        // Set wall clock + CPUTime for processing gridlet
        rgl.getGridlet().setExecParam(WallClockTime, CPUTime);

        // Send processed Gridlet back to the broker
        super.send( output, GridSimTags.SCHEDULE_NOW,
                GridSimTags.GRIDLET_RETURN, new IO_data(rgl.getGridlet(),
                    rgl.getGridlet().getGridletOutputSize(),
                    rgl.getGridlet().getUserID()) );

        // MOVE Gridlet to finished list
        gridletInExecList_.move(rgl, gridletFinishList_);

        // Schedule Next Event
        if (resource_.getResourceAllocationPolicy() ==
                ResourceCharacteristics.TIME_SHARED)
        {
            timeShared_ForecastAndScheduleNextCompletionEvent();
        }
        else
        {
            // SPACE SHARED:
            // Set PE on which Gridlet finished to FREE
            MachineList myMachineList = resource_.getMachineList();
            Machine myMachine = myMachineList.getMachine( rgl.getMachineID() );
            PEList myPEList = myMachine.getPEList();
            PE myPE = (PE) myPEList.get( rgl.getPEID() );
            myPE.setStatusFree();

            // if there are Gridlets in the QUEUE, then allocate a
            // PE to them and set Event.
            if (gridletQueueList_.size() > 0)
            {
                // choose a gridlet form the Q depending on policy and
                // assign to suitable free PE.
                // in this case, the first Gridlet in Q is choose since it
                // follows FCFS (first come first served approach.
                ResGridlet rglfromQ = (ResGridlet) gridletQueueList_.get(0);

                // remove first Gridlet from Q since it is scheduled
                // for execution.
                gridletQueueList_.remove(0);
                Gridlet gl = rglfromQ.getGridlet();
                spaceShared_AllocatePEtoGridlet(gl);
            }
        }
    }

    /**
     * Places request to raise hourly event to itself.
     * This is raised with an assumption that average local load changes
     * hourly.
     * @pre $none
     * @post $none
     */
    private void hourlyInterrupt()
    {
        if (resource_.getResourceAllocationPolicy() ==
                ResourceCharacteristics.TIME_SHARED)
        {
            // Updating So far Gridlets processed for Hourly interrupt is VALID
            // only for  TIME shared resources.
            // Because Gridlets completion time/share in SPACE shared resources
            // does not change once scheduled.
            timeShared_UpdateGridletProcessingSoFar();
            timeShared_ForecastAndScheduleNextCompletionEvent();
        }

        // Schedule Next Hour interrupt.
        super.send(super.get_id(), 60.0*60.0, GridSimTags.INSIGNIFICANT);
    }

} // end class

