/*
 * Title:        GridSim Toolkit
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation
 *               of Parallel and Distributed Systems such as Clusters and Grids
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 *
 * $Id: GridSim.java,v 1.12 2003/06/30 05:21:01 anthony Exp $
 */

package gridsim;

import java.io.*;
import java.util.*;
import eduni.simjava.*;

/**
 * The main class of the GridSim package that must be extended by GridSim
 * entities. It inherits event management and threaded entity features from
 * the <b>eduni.simjava.Sim_entity</b> class. This class adds networking and
 * event delivery features, which allow synchronous or asynchronous
 * communication for service access or delivery.
 * <p>
 * All classes that extend this class must implement a method called
 * <tt>body()</tt>, which is automatically invoked since it is expected to be
 * responsible for simulating entity behavior.
 * <p>
 * The entities that extend this class can be instantiated <i>with</i> or
 * <i>without</i> networked I/O ports. A networked GridSim entity gains
 * communication capability via the objects of GridSim's I/O entity classes,
 * <b>gridsim.Input</b> and <b>gridsim.Output</b> classes. Each I/O entity
 * will have a unique name assuming each GridSim entity that the user creates
 * has a unique name. For example, a resource entity with the name
 * <b>`Resource2'</b>,
 * will have an input entity whose name is prefixed with <b>`Input_'</b>,
 * making the input entity's full name <b>`Input_Resource2'</b>, which is
 * expected to be unique. The I/O entities are concurrent entities, but they
 * are visible within the GridSim entity and are able to communicate with
 * other GridSim entities by sending messages.
 * <p>
 * This class supports methods for simulation initilization, management and
 * flow control. The GridSim environment must be initialized to set-up the
 * simulation environment before creating any other GridSim entities at the
 * user level. This method also prepares the system for simulation by creating
 * three GridSim internal entities - <b>gridsim.GridInformationService</b>,
 * <b>gridsim.GridSimShutdown</b>,<b>gridsim.GridStatistics</b>. Invoking the
 * <tt>startGridSimulation()</tt> method starts the Grid simulation. 
 * All the resource and
 * user entities must be instantiated in between invoking the above two
 * methods.
 * <p>
 * This class supports static methods for sending and receiving messages
 * between entities directly or via network entities, managing and accessing
 * handles to various GridSim core entities, and recording statistics.
 *
 * @author       Manzur Murshed and Rajkumar Buyya
 * @version      2.1, June 2003
 * @see eduni.simjava.Sim_entity
 * @see gridsim.Output
 * @see gridsim.Input
 * @see gridsim.GridInformationService
 * @see gridsim.GridSimShutdown
 * @see gridsim.GridStatistics
 * @invariant $none
 */
public class GridSim extends Sim_entity
{
    // false means NOT invoked
    private final static int NOT_FOUND = -1;
    private boolean terminateIOEntitiesFlag_ = false;
    private double baudRate_;        // communication speed
    private boolean networkedFlag_;  // true, if networked entity, other false

    // Entity with Network Communication
    /** Reading data received via input port */
    protected Sim_port input;

    /** Sending data via output port to external entities */
    protected Sim_port output;


    /**
     * Allocates a new GridSim object
     * <b>without</b> NETWORK communication channels: "input" and
     * "output" Sim_port.
     * @param name       the name to be associated with this entity (as
     *                   required by Sim_entity class from simjava package)
     * @throws Exception This happens when creating this entity before
     *                   initializing GridSim package or this entity name is
     *                   <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @see eduni.simjava.Sim_entity
     * @pre name != null
     * @post $none
     */
    public GridSim(String name) throws Exception
    {
        super(name);
        baudRate_ = 0.0;
        networkedFlag_ = false;
        input = null;
        output = null;
    }

    /**
     * Allocates a new GridSim object
     * <b>with</b> NETWORK communication channels: "input" and
     * "output" Sim_port. In addition, this method will create <tt>Input</tt>
     * and <tt>Output</tt> object.
     * @param name       the name to be associated with this entity (as
     *                   required by Sim_entity class from simjava package)
     * @param baudRate  communication speed
     * @throws Exception This happens when creating this entity before
     *                   initializing GridSim package or this entity name is
     *                   <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @see eduni.simjava.Sim_entity
     * @see gridsim.Input
     * @see gridsim.Output
     * @pre name != null
     * @pre baudRate > 0.0
     * @post $none
     */
    public GridSim(String name, double baudRate) throws Exception
    {
        super(name);
        networkedFlag_ = true;
        this.baudRate_ = baudRate;

        input = new Sim_port("input");
        output = new Sim_port("output");

        add_port(input);
        add_port(output);

        // Every GridSim_entity_with_network has its own input/output channels.
        new Input("Input_" + name, baudRate_);

        // Connect this entity "input" port to its input's buffer's "in_port")
        Sim_system.link_ports(name, "input", "Input_" + name, "input_buffer");

        new Output("Output_" + name, baudRate_);
        Sim_system.link_ports(name, "output", "Output_"+name, "output_buffer");
    }

    /**
     * It terminates Entities managing NETWORK communication channels.
     * It can be invoked explicity to shutdown NETWORK communication channels.
     * It is advisable for all entities extending GridSim class, explicitly
     * invoke this method to terminate <tt>Input</tt> and
     * <tt>Output</tt> entities created
     * by the constructor: {@link GridSim#GridSim(String, double)}
     * @deprecated As of GridSim 2.1, replaced by
     *             {@link #terminateIOEntities()}
     * @pre $none
     * @post $none
     */
    protected void TerminateInputOutputEntities() {
        this.terminateIOEntities();
    }

    /**
     * It terminates Entities managing NETWORK communication channels.
     * It can be invoked explicity to shutdown NETWORK communication channels.
     * It is advisable for all entities extending GridSim class, explicitly
     * invoke this method to terminate <tt>Input</tt> and
     * <tt>Output</tt> entities created
     * by the constructor: {@link GridSim#GridSim(String, double)}
     * @pre $none
     * @post $none
     */
    protected void terminateIOEntities()
    {
        // If it is Networked entity and Not yet terminated, then terminate.
        if ( isNetworked() && !terminateIOEntitiesFlag_ )
        {
            // Send END_OF_SIMULATION to Input entity
            send(input, GridSimTags.SCHEDULE_NOW,
                    GridSimTags.END_OF_SIMULATION);

            // Send END_OF_SIMULATION to Output entity
            send(output, GridSimTags.SCHEDULE_NOW,
                    GridSimTags.END_OF_SIMULATION);

            terminateIOEntitiesFlag_ = true;
        }
    }

    /**
     * It terminates Entities managing NETWORK communication channels
     * @deprecated As of GridSim 2.1, replaced by
     *      {@link #finalizeGridSimulation()}
     * @see gridsim.GridSim#terminateIOEntities()
     * @pre $none
     * @post $none
     */
    protected void finalize() {
        this.finalizeGridSimulation();
    }

    /**
     * It terminates Entities managing NETWORK communication channels
     * @see gridsim.GridSim#terminateIOEntities()
     * @pre $none
     * @post $none
     */
    protected void finalizeGridSimulation() {
        terminateIOEntities();
    }

    /**
     * Check type of entity
     * @return true if entity has NETWORK communication channel, otherwise
     *          it returns false
     * @deprecated As of GridSim 2.1, replaced by {@link #isNetworked()}
     * @pre $none
     * @post $result == true || false
     */
    public boolean IsNetworked() {
        return this.isNetworked();
    }

    /**
     * Check type of entity
     * @return true if entity has NETWORK communication channel, otherwise
     *          it returns false
     * @pre $none
     * @post $result == true || false
     */
    public boolean isNetworked() {
        return networkedFlag_;
    }


    //////////// SIMULATION CONTROL METHODS START ////////////

    // Static attributes
    /**
     * Simulation start date with respect to GMT 0. This object is initialized
     * during the call to {@link #init(int, Calendar, boolean, String[],
     *             String[], String)}
     */
    public static Date SimulationStartDate = null;

    /**
     * A Random object. This object is initialized during the call to
     * {@link #init(int, Calendar, boolean, String[], String[], String)}
     */
    public static GridSimRandom rand = null;


    /**
     * Gets simulation start date with respect to GMT 0. If the return object is
     * <tt>null</tt>, then need to initialize it by calling
     * {@link #init(int, Calendar, boolean, String[], String[], String)}
     * @return an object of Date or <tt>null</tt> if it is empty
     * @deprecated As of GridSim 2.1, replaced by
     *             {@link #getSimulationStartDate()}
     * @see java.util.Date
     * @pre $none
     * @post $none
     */
    public static Date GetSimulationStartDate() {
        return getSimulationStartDate();
    }

    /**
     * Gets simulation start date with respect to GMT 0. If the return object
     * is <tt>null</tt>, then need to initialize it by calling
     * {@link #init(int, Calendar, boolean, String[], String[], String)}
     * @return an object of Date
     * @see java.util.Date
     * @pre $none
     * @post $none
     */
    public static Date getSimulationStartDate() {
        return SimulationStartDate;
    }

    /**
     * Initializes GridSim Parameters.
     * This should be called before creating any entities.
     * @param numUser  the number of User Entities created.
     *                 This parameters indicates that
     *                 <b>gridsim.GridSimShutdown</b> first waits for
     *                 User Entities's END_OF_SIMULATION signal before
     *                 issuing terminate signal to other entities
     * @param cal          the Calendar object for storing peaktime, holiday,
     *                     etc
     * @param traceFlag    true if GridSim trace need to be written
     * @param excludeFromFile  an array of String containing list of files to
     *                         be excluded from statistics
     * @param excludeFromProcessing   an array of String containing list of
     *                                processings to be excluded from writing
     *                                into a file
     * @param reportWriterName     the name of the entities which is signaled
     *                             to write any specific report just before
     *                             termination of the simulation
     * @deprecated As of GridSim 2.1, replaced by
     *             {@link #init(int, Calendar, boolean, String[],
     *             String[], String)}
     * @see gridsim.GridSimShutdown
     * @pre numUser >= 0
     * @post $none
     */
    public static void Init(int numUser, Calendar cal, boolean traceFlag,
            String[] excludeFromFile, String[] excludeFromProcessing,
            String reportWriterName)
    {
        init(numUser, cal, traceFlag, excludeFromFile,
                excludeFromProcessing, reportWriterName);
    }

    /**
     * Initializes GridSim Parameters.
     * This should be called before creating any entities.
     * @param numUser  the number of User Entities created.
     *                 This parameters indicates that
     *                 <b>gridsim.GridSimShutdown</b> first waits for
     *                 User Entities's END_OF_SIMULATION signal before
     *                 issuing terminate signal to other entities
     * @param cal          the Calendar object for storing peaktime, holiday,
     *                     etc
     * @param traceFlag    true if GridSim trace need to be written
     * @param excludeFromFile  an array of String containing list of files to
     *                         be excluded from statistics
     * @param excludeFromProcessing   an array of String containing list of
     *                                processings to be excluded from writing
     *                                into a file
     * @param reportWriterName     the name of the entities which is signaled
     *                             to write any specific report just before
     *                             termination of the simulation
     * @see gridsim.GridSimShutdown
     * @pre numUser >= 0
     * @post $none
     */
    public static void init(int numUser, Calendar cal, boolean traceFlag,
            String[] excludeFromFile, String[] excludeFromProcessing,
            String reportWriterName)
    {
        Sim_system.initialise();
        Sim_system.set_auto_trace(traceFlag);
        Sim_system.set_trc_level(1);

        SimulationStartDate = cal.getTime();
        rand = new GridSimRandom();

        try
        {
            new GridStatistics("GridStatistics", "stat.txt", true,
                    excludeFromFile, excludeFromProcessing);

            new GridSimShutdown("GridSimShutdown", numUser, reportWriterName);
            new GridInformationService("GridInformationService",
                    GridSimTags.DEFAULT_BAUD_RATE);
        }
        catch (Exception e)
        {
            e.printStackTrace();
            System.out.println("GridSim.init() : Unwanted errors happen");
        }
    }

    /**
     * Starts the execution of GridSim simulation.
     * It waits for complete execution of all entitities, i.e. until
     * all entities threads reach non-RUNNABLE state by exiting from the
     * <tt>body()</tt> method. Then, it kills threads of all entities.
     * <p>
     * <b>Note</b>: This method should be called after all the entities
     *              have been setup and added, and their ports are linked.
     * @deprecated As of GridSim 2.1, replaced by {@link #startGridSimulation()}
     * @throws NullPointerException This happens when creating this entity
     *               before initializing GridSim package or this entity name is
     *               <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @pre $none
     * @post $none
     */
    public static void Start() throws NullPointerException {
        startGridSimulation();
    }

    /**
     * Starts the execution of GridSim simulation.
     * It waits for complete execution of all entitities, i.e. until
     * all entities threads reach non-RUNNABLE state by exiting from the
     * <tt>body()</tt> method. Then, it kills threads of all entities.
     * <p>
     * <b>Note</b>: This method should be called after all the entities
     *              have been setup and added, and their ports are linked.
     * @throws NullPointerException This happens when creating this entity
     *              before initializing GridSim package or this entity name is
     *              <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @pre $none
     * @post $none
     */
    public static void startGridSimulation() throws NullPointerException
    {
        System.out.println("GridSim 2.1 --  R. Buyya and M. Murshed");
        try {
            Sim_system.run();
        }
        catch (Exception e)
        {
            throw new NullPointerException("GridSim.startGridSimulation() :" +
                    " Error - you haven't initialized GridSim.");
        }
    }

    /**
     * Gets the current simulation time (based on Sim_system.clock())
     * @return The current simulation time from the simulation clock
     * @deprecated As of GridSim 2.1, replaced by {@link #clock()}
     * @see eduni.simjava.Sim_system#clock()
     * @pre $none
     * @post $result >= 0.0
     */
    public static double Clock() {
        return clock();
    }

    /**
     * Gets the current simulation time (based on Sim_system.clock())
     * @return The current simulation time from the simulation clock
     * @see eduni.simjava.Sim_system#clock()
     * @pre $none
     * @post $result >= 0.0
     */
    public static double clock() {
        return Sim_system.clock();
    }

    /**
     * Causes the entity to hold for <tt>duration</tt> units of simulation time
     * @param duration the amount of time to hold
     * @deprecated As of GridSim 2.1, replaced by {@link #gridSimHold(double)}
     * @pre $none
     * @post $none
     */
    public void GridSimHold(double duration) {
        this.gridSimHold(duration);
    }

    /**
     * Causes the entity to hold for <tt>duration</tt> units of simulation time
     * @param duration the amount of time to hold
     * @pre $none
     * @post $none
     */
    public void gridSimHold(double duration) 
    {
        // if duration is -ve, then no use to hold -ve time.
        if (duration < 0.0) {
            return;
        }

        super.sim_hold(duration);
    }

    /**
     * Stops Grid Simulation (based on Sim_system.run_stop()).
     * This should be ony called if any of the user defined entities
     * <b>explicitly</b> want
     * to terminate simulation during execution.
     * @see eduni.simjava.Sim_system#run_stop()
     * @deprecated As of GridSim 2.1, replaced by {@link #stopGridSimulation()}
     * @throws NullPointerException This happens when creating this entity
     *               before initializing GridSim package or this entity name is
     *               <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @pre $none
     * @post $none
     */
    public static void Stop() throws NullPointerException {
        stopGridSimulation();
    }

    /**
     * Stops Grid Simulation (based on Sim_system.run_stop()).
     * This should be ony called if any of the user defined entities
     * <b>explicitly</b> want
     * to terminate simulation during execution.
     * @throws NullPointerException This happens when creating this entity
     *                before initializing GridSim package or this entity name is
     *                <tt>null</tt> or empty
     * @see gridsim.GridSim#init(int, Calendar, boolean, String[], String[],
     *          String)
     * @see eduni.simjava.Sim_system#run_stop()
     * @pre $none
     * @post $none
     */
    public static void stopGridSimulation() throws NullPointerException
    {
        try {
            Sim_system.run_stop();
        }
        catch (Exception e)
        {
            throw new NullPointerException("GridSim.stopGridSimulation() : " +
                    "Error - can't stop Grid Simulation.");
        }

    }

    ////////////  SIMULATION CONTROL METHODS END ////////////


    ////////////  ENTITY COMMUNICATION METHODS START ////////////

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @deprecated As of GridSim 2.1, replaced by {@link #send(String, double,
     *             int)}
     * @pre entityName != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void Send(String entityName, double delay, int gridSimTag) {
        this.send(entityName, delay, gridSimTag);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @pre entityName != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void send(String entityName, double delay, int gridSimTag)
    {
        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(GridSim.getEntityId(entityName), delay, gridSimTag);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @deprecated As of GridSim 2.1, replaced by {@link #send(String, double,
     *             int, Object)}
     * @pre entityName != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void Send(String entityName, double delay, int gridSimTag,
            Object data)
    {
        this.send(entityName, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityName the name of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @pre entityName != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void send(String entityName, double delay, int gridSimTag,
            Object data)
    {
        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(GridSim.getEntityId(entityName), delay, gridSimTag,
                data);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @deprecated As of GridSim 2.1, replaced by {@link #send(int, double,
     *             int)}
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @post $none
     */
    protected void Send(int entityID, double delay, int gridSimTag) {
        this.send(entityID, delay, gridSimTag);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @post $none
     */
    protected void send(int entityID, double delay, int gridSimTag)
    {
        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(entityID, delay, gridSimTag);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @deprecated As of GridSim 2.1, replaced by {@link #send(int, double,
     *             int, Object)}
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void Send(int entityID, double delay, int gridSimTag, Object data)
    {
        this.send(entityID, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param entityID   the id number of the destination entity
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @pre entityID > 0
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void send(int entityID, double delay, int gridSimTag, Object data)
    {
        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(entityID, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param destPort   A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @deprecated As of GridSim 2.1, replaced by {@link #send(Sim_port, double,
     *             int)}
     * @pre destPort != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void Send(Sim_port destPort, double delay, int gridSimTag) {
        this.send(destPort, delay, gridSimTag);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param destPort   A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @pre destPort != null
     * @pre delay >= 0.0
     * @post $none
     */
    protected void send(Sim_port destPort, double delay, int gridSimTag)
    {
        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(destPort, delay, gridSimTag);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param destPort  A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @deprecated As of GridSim 2.1, replaced by {@link #send(Sim_port, double,
     *             int, Object)}
     * @pre destPort != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void Send(Sim_port destPort, double delay, int gridSimTag,
            Object data)
    {
        this.send(destPort, delay, gridSimTag, data);
    }

    /**
     * Sends an event/message to Entity by 'delaying' simulation time
     * from now with user-defined tag representing the type of event.
     * @param destPort  A reference to the port to send the event out of
     * @param delay      how long the event/message should be delivered
     *                   from the current simulation time
     * @param gridSimTag an user-defined number representing the type of
     *                   an event/message
     * @param data       A reference to data to be sent with the event
     * @pre destPort != null
     * @pre delay >= 0.0
     * @pre data != null
     * @post $none
     */
    protected void send(Sim_port destPort, double delay, int gridSimTag,
            Object data)
    {
        // if delay is -ve, then it doesn't make sense. So resets to 0.0
        if (delay < 0.0) {
            delay = 0.0;
        }

        super.sim_schedule(destPort, delay, gridSimTag, data);
    }

    /**
     * Gets the data passed in this event
     * @return A reference to the data
     * @deprecated As of GridSim 2.1, replaced by {@link #receiveEventObject()}
     * @pre $none
     * @post $result != null
     */
    protected Object ReceiveEventObject() {
        return this.receiveEventObject();
    }

    /**
     * Gets the data passed in this event
     * @return A reference to the data
     * @pre $none
     * @post $result != null
     */
    protected Object receiveEventObject()
    {
        Sim_event ev = new Sim_event();
        super.sim_get_next(ev);
        return ev.get_data();
    }

    /**
     * Gets the data passed in this event
     * @param sourcePort a Sim_port object which is used to connect entities
     *                    for  event passing
     * @return A reference to the data
     * @see eduni.simjava.Sim_port
     * @pre sourcePort != null
     * @post $result != null
     */
    protected Object ReceiveEventObject(Sim_port sourcePort) {
        return this.receiveEventObject(sourcePort);
    }

    /**
     * Gets the data passed in this event
     * @param sourcePort a Sim_port object which is used to connect entities
     *                    for  event passing
     * @return A reference to the data
     * @see eduni.simjava.Sim_port
     * @pre sourcePort != null
     * @post $result != null
     */
    protected Object receiveEventObject(Sim_port sourcePort)
    {
        Sim_event ev = new Sim_event();
        super.sim_get_next( new Sim_from_port(sourcePort), ev );
        return ev.get_data();
    }

    /**
     * Sends a gridlet based on its resource ID to Entity
     * @param gl            a Gridlet object to be sent
     * @param resourceID    a unique resource ID
     * @deprecated As of GridSim 2.1, replaced by {@link #gridletSubmit(Gridlet,
     *             int)}
     * @pre gl != null
     * @pre resourceID > 0
     * @post $none
     */
    protected void GridletSubmit(Gridlet gl, int resourceID) {
        this.gridletSubmit(gl, resourceID);
    }

    /**
     * Sends a gridlet based on its resource ID to Entity
     * @param gl            a Gridlet object to be sent
     * @param resourceID    a unique resource ID
     * @pre gl != null
     * @pre resourceID > 0
     * @post $none
     */
    protected void gridletSubmit(Gridlet gl, int resourceID)
    {
        try {
            gl.setGridletStatus(Gridlet.QUEUED);
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        send( output, GridSimTags.SCHEDULE_NOW, GridSimTags.GRIDLET_SUBMIT,
                new IO_data(gl, gl.getGridletFileSize(), resourceID) );
    }

    /**
     * Gets a Gridlet object passed in this event
     * @return A reference to Gridlet object
     * @deprecated As of GridSim 2.1, replaced by {@link #gridletReceive()}
     * @pre $none
     * @post $result != null
     */
    protected Gridlet GridletReceive() {
        return this.gridletReceive();
    }

    /**
     * Gets a Gridlet object passed in this event
     * @return A reference to Gridlet object
     * @pre $none
     * @post $result != null
     */
    protected Gridlet gridletReceive()
    {
        Gridlet glFinished = (Gridlet) receiveEventObject(input);
        return glFinished;
    }

    /**
     * Cancels a Gridlet object <b>NOT YET IMPLEMENTED</b>
     * @param gl    a Gridlet object
     * @param resource_id   a grid resource id
     * @pre gl != null
     * @pre resource_id >= 0
     * @post $none
     */
    protected void GridletCancel(Gridlet gl, int resource_id )
    {
        //gl.setGridletStatus(Gridlet.QUEUED);
        //send(output, GridSimTags.SCHEDULE_NOW, GridSimTags.GRIDLET_SUBMIT,
        //          new IO_data(gl, gl.getGridletFileSize(), resource_id));
        System.out.println("GridSim.GridletCancel() : Not Yet Implemented");
    }

    ////////////  ENTITY  COMMUNICATION METHODS END ////////////


    ////////////  ENTITY ID/NAME CONVERSION METHODS START ////////

    /**
     * Gets the name of the current entity
     * @return the Entity name or <tt>null</tt> if this object does not have
     *         one
     * @deprecated As of GridSim 2.1, replaced by {@link #getEntityName()}
     * @pre $none
     * @post $none
     */
    public String GetEntityName() {
        return this.getEntityName();
    }

    /**
     * Gets the name of the current entity
     * @return the Entity name or <tt>null</tt> if this object does not have
     *         one
     * @pre $none
     * @post $none
     */
    public String getEntityName()
    {
        try {
            return super.get_name();
        }
        catch (Exception e) {
            return null;
        }
    }

    /**
     * Gets name of the entity given its entity ID
     * @param entityID  the entity ID
     * @return the Entity name or <tt>null</tt> if this object does not have
     *         one
     * @deprecated As of GridSim 2.1, replaced by {@link #getEntityName(int)}
     * @pre entityID > 0
     * @post $none
     */
    public static String GetEntityName(int entityID) {
        return getEntityName(entityID);
    }

    /**
     * Gets name of the entity given its entity ID
     * @param entityID  the entity ID
     * @return the Entity name or <tt>null</tt> if this object does not have
     *         one
     * @pre entityID > 0
     * @post $none
     */
    public static String getEntityName(int entityID)
    {
        try {
            return Sim_system.get_entity(entityID).get_name();
        }
        catch (Exception e) {
            return null;
        }
    }

    /**
     * Gets ID of the Entity given its name
     * @param entityName    an Entity name
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @deprecated As of GridSim 2.1, replaced by {@link #getEntityId(String)}
     * @pre entityName != null
     * @post $result >= -1
     */
    public static int GetEntityId(String entityName) {
        return getEntityId(entityName);
    }

    /**
     * Gets ID of the Entity given its name
     * @param entityName    an Entity name
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @pre entityName != null
     * @post $result >= -1
     */
    public static int getEntityId(String entityName)
    {
        try {
            return (int) Sim_system.get_entity_id(entityName);
        }
        catch (Exception e) {
            return NOT_FOUND;
        }
    }

    /**
     * Gets ID of Entity named "GridStatistics"
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #getGridStatisticsEntityId()}
     * @pre $none
     * @post $result >= -1
     */
    public static int GridStatisticsEntityId() {
        return getGridStatisticsEntityId();
    }

    /**
     * Gets ID of Entity named "GridStatistics"
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @pre $none
     * @post $result >= -1
     */
    public static int getGridStatisticsEntityId()
    {
        try {
            return (int) Sim_system.get_entity_id("GridStatistics");
        }
        catch (Exception e) {
            return NOT_FOUND;
        }
    }

    /**
     * Gets the ID of Entity named "GridInformationService"
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #getGridInfoServiceEntityId()}
     * @pre $none
     * @post $result >= -1
     */
    public static int GridInformationServiceEntityId() {
        return getGridInfoServiceEntityId();
    }

    /**
     * Gets the ID of Entity named "GridInformationService"
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @pre $none
     * @post $result >= -1
     */
    public static int getGridInfoServiceEntityId()
    {
        try {
            return (int) Sim_system.get_entity_id("GridInformationService");
        }
        catch (Exception e) {
            return NOT_FOUND;
        }
    }

    /**
     * Gets ID of Entity named "GridSimShutdown"
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #getGridSimShutdownEntityId()}
     * @pre $none
     * @post $result >= -1
     */
    public static int GridSimShutdownEntityId() {
        return getGridSimShutdownEntityId();
    }

    /**
     * Gets ID of Entity named "GridSimShutdown"
     * @return the Entity ID or <tt>-1</tt> if it is not found
     * @pre $none
     * @post $result >= -1
     */
    public static int getGridSimShutdownEntityId()
    {
        try {
            return (int) Sim_system.get_entity_id("GridSimShutdown");
        }
        catch (Exception e) {
            return NOT_FOUND;
        }
    }

    /**
     * Tells all user entities to shut down the simulation.
     * <b>gridsim.GridSimShutdown</b> entity waits for all users
     * termination before shuting down other entities.
     * @see gridsim.GridSimShutdown
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #shutdownUserEntity()}
     * @pre $none
     * @post $none
     */
    protected void ShutdownUserEntity() {
        this.shutdownUserEntity();
    }

    /**
     * Tells all user entities to shut down the simulation.
     * <b>gridsim.GridSimShutdown</b> entity waits for all users
     * termination before shuting down other entities.
     * @see gridsim.GridSimShutdown
     * @pre $none
     * @post $none
     */
    protected void shutdownUserEntity()
    {
        send(GridSim.getGridSimShutdownEntityId(), GridSimTags.SCHEDULE_NOW,
                GridSimTags.END_OF_SIMULATION);
    }

    /**
     * Tells all user entities that defined <tt>ReportWriter</tt> to shut
     * down the simulation.
     * If a <tt>ReportWriter</tt> is <b>NOT</b> defined (i.e. null) during
     * <tt>GridSim.Init()</tt>,
     * then <b>gridsim.GridSimShutdown</b> will invoke it.
     * @see gridsim.GridSimShutdown
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #shutdownGridStatisticsEntity()}
     * @pre $none
     * @post $none
     */
    protected void ShutdownGridStatisticsEntity() {
        this.shutdownGridStatisticsEntity();
    }

    /**
     * Tells all user entities that defined <tt>ReportWriter</tt> to shut
     * down the simulation.
     * If a <tt>ReportWriter</tt> is <b>NOT</b> defined (i.e. null) during
     * <tt>GridSim.Init()</tt>,
     * then <b>gridsim.GridSimShutdown</b> will invoke it.
     * @see gridsim.GridSimShutdown
     * @pre $none
     * @post $none
     */
    protected void shutdownGridStatisticsEntity()
    {
        send(GridSim.getGridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
                GridSimTags.END_OF_SIMULATION);
    }

    /**
     * Sends a request to Grid Information Service (GIS) entity to get the
     * list of Grid resources
     * @return A LinkedList containing GridResource ID (as an Integer object)
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #getGridResourceList()}
     * @pre $none
     * @post $result != null
     */
    public LinkedList GetGridResourceList() {
        return this.getGridResourceList();
    }

    /**
     * Sends a request to Grid Information Service (GIS) entity to get the
     * list of Grid resources
     * @return A LinkedList containing GridResource ID (as an Integer object)
     * @pre $none
     * @post $result != null
     */
    public LinkedList getGridResourceList()
    {
        LinkedList ResIDList;

        // Requesting the GridInformationService to send a list of resources
        // to "IO_data", my entity ID, size of data transfer,
        // and destination entity ID is passed.
        send( output, GridSimTags.SCHEDULE_NOW, GridSimTags.RESOURCE_LIST,
                new IO_data(new Integer(super.get_id()), 4,
                    GridSim.getGridInfoServiceEntityId()) );

        ResIDList = (LinkedList) receiveEventObject();
        return ResIDList;
    }

    /**
     * Gets a ResourceCharacteristics object
     * @param resourceID   the resource ID
     * @return An object of gridsim.ResourceCharacteristics
     * @see gridsim.ResourceCharacteristics
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #getResourceCharacteristics(int)}
     * @pre resourceID > 0
     * @post $result != null
     */
    public ResourceCharacteristics GetResourceCharacteristics(int resourceID) {
        return this.getResourceCharacteristics(resourceID);
    }

    /**
     * Gets a ResourceCharacteristics object
     * @param resourceID   the resource ID
     * @return An object of gridsim.ResourceCharacteristics
     * @see gridsim.ResourceCharacteristics
     * @pre resourceID > 0
     * @post $result != null
     */
    public ResourceCharacteristics getResourceCharacteristics(int resourceID)
    {
        // Get Resource Characteristic Info: Send Request and Receive Event/Msg
        send( output, GridSimTags.SCHEDULE_NOW,
                GridSimTags.RESOURCE_CHARACTERISTICS,
                new IO_data(new Integer(super.get_id()), 4, resourceID) );

        ResourceCharacteristics res = (ResourceCharacteristics)
                    receiveEventObject(input);
        return res;
    }

    /**
     * Gets Resource Dynamic Information
     * @param resourceID   the resource ID
     * @return An object of Accumulator containing the resource load
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #getResourceDynamicInfo(int)}
     * @pre resourceID > 0
     * @post $result != null
     */
    public Accumulator GetResourceDynamicInformation(int resourceID) {
        return this.getResourceDynamicInfo(resourceID);
    }

    /**
     * Gets Resource Dynamic Information
     * @param resourceID   the resource ID
     * @return An object of Accumulator containing the resource load
     * @pre resourceID > 0
     * @post $result != null
     */
    public Accumulator getResourceDynamicInfo(int resourceID)
    {
        // Get Resource Dynamic information
        send( output, GridSimTags.SCHEDULE_NOW, GridSimTags.RESOURCE_DYNAMICS,
                new IO_data(new Integer(super.get_id()), 4, resourceID) );

        Accumulator accLoad = (Accumulator) receiveEventObject(input);
        return accLoad;
    }

    ////////////  ENTITY ID/NAME CONVERSION METHODS END ////////////



    ////////////  METHOD FOR RECORDING Statistics Information BEGIN ////////

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #recordStatistics(String, double)}
     * @pre category != null
     * @pre data >= 0.0
     * @post $none
     */
    public void RecordStatistics(String category, double data) {
        this.recordStatistics(category, data);
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @pre category != null
     * @pre data >= 0.0
     * @post $none
     */
    public void recordStatistics(String category, double data)
    {
        send(getGridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
                GridSimTags.RECORD_STATISTICS, new Stat(GridSim.clock(),
                    category, super.get_name(), data));
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #recordStatistics(String, int)}
     * @pre category != null
     * @pre data >= 0.0
     * @post $none
     */
    public void RecordStatistics(String category, int data) {
        this.recordStatistics(category, data);
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @pre category != null
     * @pre data >= 0.0
     * @post $none
     */
    public void recordStatistics(String category, int data)
    {
        send(getGridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
                GridSimTags.RECORD_STATISTICS, new Stat(GridSim.clock(),
                    category, super.get_name(), data));
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #recordStatistics(String, String)}
     * @pre category != null
     * @pre data != null
     * @post $none
     */
    public void RecordStatistics(String category, String data) {
        this.recordStatistics(category, data);
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @pre category != null
     * @pre data != null
     * @post $none
     */
    public void recordStatistics(String category, String data)
    {
        send(getGridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
                GridSimTags.RECORD_STATISTICS, new Stat(GridSim.clock(),
                    category, super.get_name(), data));
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @deprecated As of GridSim 2.1, replaced by {@link
     *             #recordStatistics(String, boolean)}
     * @pre category != null
     * @pre data == true || false
     * @post $none
     */
    public void RecordStatistics(String category, boolean data) {
        this.recordStatistics(category, data);
    }

    /**
     * Records statistics during the event
     * @param category  a category name
     * @param data      a value to be recorded
     * @pre category != null
     * @pre data == true || false
     * @post $none
     */
    public void recordStatistics(String category, boolean data)
    {
        send(getGridStatisticsEntityId(), GridSimTags.SCHEDULE_NOW,
                GridSimTags.RECORD_STATISTICS, new Stat(GridSim.clock(),
                    category, super.get_name(), data));
    }

} // end class

