/*
 * Title:        GridSim Toolkit
 * Description:  GridSim (Grid Simulation) Toolkit for Modeling and Simulation 
 *               of Parallel and Distributed Systems such as Clusters and Grids
 * Licence:      GPL - http://www.gnu.org/copyleft/gpl.html
 *
 * $Id: ResGridlet.java,v 1.8 2003/06/30 05:21:02 anthony Exp $
 */

package gridsim;

import java.util.Calendar;


/**
 * GridSim ResGridlet represents a Gridlet submitted to the resource for
 * processing. 
 * <p>
 * It contains a Gridlet object along with its arrival time and
 * the ID of the machine and the PE (Processing Element) allocated to it.
 * It acts as a placeholder for maintaining the amount of resource share
 * allocated at various times for simulating time-shared and space-shared 
 * scheduling using internal events.
 *
 * @author       Manzur Murshed and Rajkumar Buyya
 * @version      2.1, June 2003
 * @invariant $none
 */
public class ResGridlet
{
    private Gridlet gridlet_;
    private double arrivalTime_;
    private double finishedTime_;
    private double gridletFinishedSoFar_;

    // For SPACE shared Resources:
    private int machineID_;
    private int peID_;


    /**
     * Allocates a new ResGridlet object upon the arrival of a Gridlet object.
     * The arriving time is determined by GridSim.Clock()
     * @param gridlet a gridlet object
     * @see GridSim#Clock()
     * @pre gridlet != null
     * @post $none
     */
    public ResGridlet(Gridlet gridlet)
    {
        this.gridlet_ = gridlet;
        this.arrivalTime_ = GridSim.clock();
        this.finishedTime_ = -1;  // Cannot finish in this hourly slot.
        this.gridletFinishedSoFar_ = 0.0;
        this.machineID_ = -1;
        this.peID_ = -1;
    }

    /**
     * Set the machine and PE (Processing Element) ID
     * @param machineID   machine ID
     * @param peID        PE ID
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #setMachineAndPEID(int, int)}
     * @pre machineID >= 0
     * @pre peID >= 0
     * @post $none
     */
    public void SetIDs(int machineID, int peID) {
        this.setMachineAndPEID(machineID, peID);
    }
    
    /**
     * Set the machine and PE (Processing Element) ID
     * @param machineID   machine ID
     * @param peID        PE ID
     * @pre machineID >= 0
     * @pre peID >= 0
     * @post $none
     */
    public void setMachineAndPEID(int machineID, int peID)
    {
        this.machineID_ = machineID;
        this.peID_ = peID;
    }

    /**
     * Get machine ID
     * @return machine ID or <tt>-1</tt> if it is not specified before
     * @deprecated As of GridSim 2.1, replaced by {@link #getMachineID()}
     * @pre $none
     * @post $result >= -1
     */
    public int GetMachineID() {
        return this.getMachineID();
    }
    
    /**
     * Get machine ID
     * @return machine ID or <tt>-1</tt> if it is not specified before
     * @pre $none
     * @post $result >= -1
     */
    public int getMachineID() {
        return machineID_;
    }

    /**
     * Get PE ID
     * @return PE ID or <tt>-1</tt> if it is not specified before
     * @deprecated As of GridSim 2.1, replaced by {@link #getPEID()}
     * @pre $none
     * @post $result >= -1
     */
    public int GetPEID() {
        return this.getPEID();
    }
    
    /**
     * Get PE ID
     * @return PE ID or <tt>-1</tt> if it is not specified before
     * @pre $none
     * @post $result >= -1
     */
    public int getPEID() {
        return peID_;
    }

    /**
     * Get the remaining gridlet length 
     * @return gridlet length
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #getRemainingGridletLength()}
     * @pre $none
     * @post $result >= 0
     */
    public double GetRemainingLength() {
        return this.getRemainingGridletLength();
    }
    
    /**
     * Get the remaining gridlet length 
     * @return gridlet length
     * @pre $none
     * @post $result >= 0
     */
    public double getRemainingGridletLength() {
        return gridlet_.getGridletLength() - gridletFinishedSoFar_;
    }

    /**
     * A method that updates the length of gridlet that has been completed  
     * @param miLength gridlet length in Million Instructions (MI)
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #updateGridletFinishedSoFar(double)}
     * @pre miLength >= 0.0
     * @post $none
     */
    public void UpdateGridletFinishedSoFar(double miLength) {
        this.updateGridletFinishedSoFar(miLength);
    }
    
    /**
     * A method that updates the length of gridlet that has been completed  
     * @param miLength gridlet length in Million Instructions (MI)
     * @pre miLength >= 0.0
     * @post $none
     */
    public void updateGridletFinishedSoFar(double miLength) {
        gridletFinishedSoFar_ += miLength;
    }

    /**
     * Get arrival time of a gridlet
     * @return arrival time
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #getGridletArrivalTime()}
     * @pre $none
     * @post $result >= 0.0
     */
    public double GetArrivalTime() {
        return this.getGridletArrivalTime();
    }
    
    /**
     * Get arrival time of a gridlet
     * @return arrival time
     * @pre $none
     * @post $result >= 0.0
     */
    public double getGridletArrivalTime() {
        return arrivalTime_;
    }

    /**
     * Determine Time by which this Gridlet can finish.
     * <br>
     * <b>NOTE:</b> To be used by Time shared resources only.
     * @param availableRating  the available MIPS considering current hour 
     *                         load
     * @param rCalendar        Resourcce Calendar
     * @throws NullPointerException  if ResourceCalendar object is <tt>null</tt>
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #setFinishTimeForTimeSharedResource(double, 
     *             ResourceCalendar)}
     * @pre availableRating >= 0.0
     * @pre rCalendar != null
     * @post $none
     */
    public void SetFinishTimeForTimeSharedResource(double availableRating, 
                    ResourceCalendar rCalendar) throws NullPointerException
    {
        this.setFinishTimeForTimeSharedResource(availableRating, rCalendar);
    }
    
    /**
     * Determine Time by which this Gridlet can finish.
     * <br>
     * <b>NOTE:</b> To be used by Time shared resources only.
     * @param availableRating  the available MIPS considering current hour 
     *                         load
     * @param rCalendar        Resourcce Calendar
     * @throws NullPointerException  if ResourceCalendar object is <tt>null</tt>
     * @pre availableRating >= 0.0
     * @pre rCalendar != null
     * @post $none
     */
    public void setFinishTimeForTimeSharedResource(double availableRating, 
                    ResourceCalendar rCalendar) throws NullPointerException
    {
        if (rCalendar == null) {
            throw new NullPointerException(
                    "Gridlet.setFinishTimeForTimeSharedResource() : Error - " +
                    "ResourceCalendar object must not be null or empty.");
        }
        
        double fTime = (getRemainingGridletLength() / availableRating) 
                       + GridSim.clock();
        
        Calendar finish = rCalendar.getCalendarAtSimulationTime(fTime);
        Calendar NextHour = rCalendar.getCurrentCalendar();
        
        NextHour.add(Calendar.HOUR, 1);
        NextHour.set(Calendar.MINUTE, 0);
        NextHour.set(Calendar.SECOND, 0);

        if ( finish.before(NextHour) ) {
            finishedTime_ = fTime;
        }

        // it finishes after the Hourly interrupt and AvailableMIPSRating 
        // will change,
        // so current predication may not be correct.
        else {
            finishedTime_ = -1; 
        }
    }

    /**
     * Determine Time by which this Gridlet can finish.
     * <br>
     * <b>NOTE:</b> To be used by Space shared resources only.
     * @param availableRating  this is 100% availability of PE MIPS since 
     *                         it runs in dedicated mode
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #setFinishTimeForSpaceSharedResource(double)}
     * @pre availableRating >= 0.0
     * @post $none
     */
    public void SetFinishTimeForSpaceSharedResource(double availableRating)
    {
        this.setFinishTimeForSpaceSharedResource(availableRating);
    }
    
    /**
     * Determine Time by which this Gridlet can finish.
     * <br>
     * <b>NOTE:</b> To be used by Space shared resources only.
     * @param availableRating  this is 100% availability of PE MIPS since 
     *                         it runs in dedicated mode
     * @pre availableRating >= 0.0
     * @post $none
     */
    public void setFinishTimeForSpaceSharedResource(double availableRating)
    {
        finishedTime_ = (getRemainingGridletLength() / availableRating) 
                        + GridSim.clock();
    }

    /**
     * Get the Gridlet's finish time
     * @return finish time of a gridlet or <tt>-1.0</tt> if 
     *         it cannot finish in this hourly slot
     * @deprecated As of GridSim 2.1, replaced by 
     *             {@link #getGridletFinishTime()}
     * @pre $none
     * @post $result >= -1.0
     */
    public double GetFinishTime() {
        return this.getGridletFinishTime();
    }
    
    /**
     * Get the Gridlet's finish time
     * @return finish time of a gridlet or <tt>-1.0</tt> if 
     *         it cannot finish in this hourly slot
     * @pre $none
     * @post $result >= -1.0
     */
    public double getGridletFinishTime() {
        return finishedTime_;
    }

    /**
     * Get this Gridlet object
     * @return gridlet object
     * @deprecated As of GridSim 2.1, replaced by {@link #getGridlet()}
     * @pre $none
     * @post $result != null
     */
    public Gridlet GetGridlet() {
        return this.getGridlet();
    }
    
    /**
     * Get this Gridlet object
     * @return gridlet object
     * @pre $none
     * @post $result != null
     */
    public Gridlet getGridlet() {
        return gridlet_;
    }
    
} // end class

